package VISD::SysInfo; # $Id: SysInfo.pm 104 2020-10-05 13:13:46Z abalama $
use strict;

=encoding utf8

=head1 NAME

VISD::SysInfo - Getting system information

=head1 VERSION

Version 1.02

=head1 SYNOPSIS

    use VISD::SysInfo;
    my %sysinfo = get_sysinfo();

=head1 DESCRIPTION

Getting system information

=head2 df

    my %df = VISD::SysInfo::df();
    my %df = VISD::SysInfo::df("/");

Returns df values for root mount point

  {
    'avail' => '105652924416',
    'free' => '130464854016',
    'total' => '487987372032',
    'used' => '357522518016',
    'bsize' => '4096'
  }

=head2 get_blkio

    my %blkio = get_blkio();

Returns BlockIO statistic (in bytes) table

  {
    '11:0' => {
      'Async' => '0',
      'Read' => '0',
      'Sync' => '0',
      'Total' => '0',
      'Write' => '0'
    },
  }

=head2 get_cpuinfo

    my %cpuinfo = get_cpuinfo();

Returns common CPU information

  {
    'cpus' => 4,
    'hvm' => 1,
    'mhz' => '2723.388',
    'model' => 'Intel(R) Core(TM) i5-2450M CPU @ 2.50GHz',
    'sockets' => 1,
    'user_hz' => 100
  }

=head2 get_disk_free

    my %disk_free = get_disk_free();

Returns disk free (df)

  {
    '/dev/sda1' => {
      'avail' => '105653059584',
      'free' => '130464989184',
      'freq' => '0',
      'fstype' => 'ext4',
      'opts' => 'rw,relatime,errors=remount-ro,data=ordered',
      'passno' => '0',
      'point' => '/',
      'total' => '487987372032',
      'used' => '357522382848',
      'diskread' => '101',
      'diskwrite' => '5'
    },
  }

=head2 get_diskstats

    my %diskstats = get_diskstats();

Returns diskstats

  {
    '/dev/dm-1' => {
      'name' => 'dm-1',
      'diskread' => '96202',
      'diskwrite' => '27880'
    },
  }

=head2 get_filesystems

    my @filesystems = get_filesystems();

Return filesystems without "nodev" filesystems

    my @filesystems = get_filesystems(1);

Return all available filesystems

  [
    'ext3',
    'ext2',
    'ext4',
    'squashfs',
    'vfat',
    'fuseblk'
  ]

=head2 get_mounts

    my %mounts = get_mounts();

Returns mount table

  {
    '/dev/sda1' => {
      'freq' => '0',
      'fstype' => 'ext4',
      'opts' => 'rw,relatime,errors=remount-ro,data=ordered',
      'passno' => '0',
      'point' => '/',
      'majmin' => '8:1',
      'id' => '84',
    },
  }

=head2 get_net_dev

    my %net_dev = get_net_dev();

Returns actual network devices information

  {
    'tun0' => {
      'receive' => '3215250',
      'transmit' => '3090378'
    },
  }

=head2 get_net_ipv6_route

    my @route_table = get_net_ipv6_route();

Returns IPv6 routing table

  [
    {
      'destination' => 'fe80:0000:0000:0000:0000:0000:0000:0000',
      'gateway' => '0000:0000:0000:0000:0000:0000:0000:0000',
      'iface' => 'tun0',
      'metric' => 256,
      'prefix' => 64
    },
  ]

=head2 get_net_route

    my @route_table = get_net_route();

Returns IPv4 routing table

  [
    {
      'destination' => '0.0.0.0',
      'gateway' => '192.168.123.1',
      'iface' => 'wlp1s0',
      'mask' => '0.0.0.0',
      'metric' => '600',
      'mtu' => '0'
    },
  ]

=head2 get_net_tcp

    my @tcp = get_net_tcp();

Returns TCP & TCP6 data table

=head2 get_net_udp

    my @udp = get_net_udp();

Returns UDP & UDP6 data table

=head2 get_stat

    my %stat = get_stat();

Returns system stat information

  {
    'cpuusage' => 0,
    'ctime' => '1584945928.5248',
    'idle' => '22995270',
    'iodelay' => 0,
    'iowait' => '123344',
    'nice' => '2391',
    'system' => '132405',
    'used' => 545043,
    'user' => '410247'
  }

=head2 get_sysinfo

    my %sysinfo = get_sysinfo();

Returns sysinfo structure as perl hash

  {
    'bufferram' => 249507840,
    'freeram' => 620646400,
    'freeswap' => 4042133504,
    'load' => [
      '0.16',
      '0.14',
      '0.09'
    ],
    'load1' => '0.1552734375',
    'load15' => '0.0947265625',
    'load5' => '0.14453125',
    'procs' => 771,
    'sharedram' => 280379392,
    'totalram' => 4050661376,
    'totalswap' => 4201639936,
    'uptime' => 59415,
    'usedram' => 3430014976,
    'usedswap' => 159506432
  }

=head2 get_uname

    my %uname = get_uname();

Get name of current operating system. See L<POSIX>

  {
    'machine' => 'x86_64',
    'nodename' => 'mnslpt',
    'release' => '4.8.0-59-generic',
    'sysname' => 'Linux',
    'version' => '#64-Ubuntu SMP Thu Jun 29 19:38:34 UTC 2017'
  }

=head2 get_uptime

    my $uptime = get_uptime();

Returns current uptime in seconds

=head1 SEE ALSO

L<VISD>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See C<LICENSE> file and L<https://dev.perl.org/licenses/>

=cut

use vars qw/$VERSION @EXPORT/;
$VERSION = '1.02';

use base qw /Exporter/;
use Carp qw /carp croak/;
use POSIX ();
use IO::File;
use Time::HiRes qw/gettimeofday/;
use File::Spec ();
use Cwd qw/abs_path/;

use XSLoader;
XSLoader::load( "VISD", $VERSION );

use constant {
    IO_DELAY_INTERVAL       => 1000,
    DEV_DIR                 => '/dev',
    PROC_CPUINFO_FILE       => '/proc/cpuinfo',
    PROC_UPTIME_FILE             => '/proc/uptime',
    PROC_STAT_FILE          => '/proc/stat',
    PROC_NET_ROUTE_FILE     => '/proc/net/route',
    PROC_NET_IPV6_ROUTE_FILE=> '/proc/net/ipv6_route',
    PROC_NET_DEV_FILE       => '/proc/net/dev',
    PROC_MOUNTS_FILE        => '/proc/self/mounts',
    PROC_MOUNTINFO_FILE     => '/proc/self/mountinfo',
    PROC_FILESYSTEMS_FILE   => '/proc/filesystems',
    PROC_DISKSTATS_FILE     => '/proc/diskstats',
    PROC_BLKIO_FILE         => '/sys/fs/cgroup/blkio/blkio.throttle.io_service_bytes',
    PROC_NET_TCP            => '/proc/net/tcp',
    PROC_NET_TCP6           => '/proc/net/tcp6',
    PROC_NET_UDP            => '/proc/net/udp',
    PROC_NET_UDP6           => '/proc/net/udp6',

    # TCP ST
    TCP_ST_NAMES            => [qw/
        UNKNOWN
        ESTABLISHED
        SYN_SENT
        SYN_RECV
        FIN_WAIT1
        FIN_WAIT2
        TIME_WAIT
        CLOSE
        CLOSE_WAIT
        LAST_ACK
        LISTEN
        CLOSING
        NEW_SYN_RECV
        MAX_STATES
    /],
};

@EXPORT = qw(
        get_sysinfo
        get_uname get_uptime
        get_cpuinfo
        get_stat
        get_net_dev get_net_route get_net_ipv6_route
        get_mounts get_filesystems get_disk_free get_diskstats get_blkio
        get_net_tcp get_net_udp
    );

my $clock_ticks = POSIX::sysconf(&POSIX::_SC_CLK_TCK);
my %last_proc_stat; # See get_stat

sub get_sysinfo {
    my $sysinfo = _getsysinfo();

    # Calculate a usedram
    $sysinfo->{usedram} = $sysinfo->{totalram} - $sysinfo->{freeram};

    # Calculate a usedswap
    $sysinfo->{usedswap} = $sysinfo->{totalswap} - $sysinfo->{freeswap};

    return %$sysinfo;
}
sub get_uname {
    my ($sysname, $nodename, $release, $version, $machine) = POSIX::uname();
    return (
        sysname => $sysname // '',
        nodename => $nodename // '',
        release => $release // '',
        version => $version // '',
        machine => $machine // '',
    );
}
sub get_uptime {
    my $uptime = 0;
    my $fh = IO::File->new(PROC_UPTIME_FILE, "r");
    return $uptime if !$fh;

    my $line = <$fh>;
    ($uptime) = $line =~ /^(\d+)/;

    $fh->close;
    return $uptime;
}
sub get_cpuinfo {
    my %res = (
        user_hz     => $clock_ticks,
        model       => 'unknown',
        mhz         => 0,
        cpus        => 1,
        sockets     => 1,
    );

    my $fh = IO::File->new(PROC_CPUINFO_FILE, "r");
    return %res if !$fh;

    my $idhash = {};
    my $count = 0;
    while (defined(my $line = <$fh>)) {
        if ($line =~ m/^processor\s*:\s*\d+\s*$/i) {
            $count++;
        } elsif ($line =~ m/^model\s+name\s*:\s*(.*)\s*$/i) {
            $res{model} = $1 if $res{model} eq 'unknown';
        } elsif ($line =~ m/^cpu\s+MHz\s*:\s*(\d+\.\d+)\s*$/i) {
            $res{mhz} = $1 if !$res{mhz};
        } elsif ($line =~ m/^flags\s*:.*(vmx|svm)/) {
            $res{hvm} = 1; # Hardware Virtual Machine (Intel VT / AMD-V)
        } elsif ($line =~ m/^physical id\s*:\s*(\d+)\s*$/i) {
            $idhash->{$1} = 1;
        }
    }
    $res{sockets} = scalar(keys %$idhash) || 1;
    $res{cpus} = $count;

    $fh->close;
    return %res;
}
sub get_stat {
    my %res = (
        user    => 0,
        nice    => 0,
        system  => 0,
        idle    => 0,
        iowait  => 0,
        used    => 0,
        iodelay => 0, # wait
        cpuusage=> 0, # cpu
    );

    # Load stat data
    my $cpucount = 0;
    if (my $fh = IO::File->new(PROC_STAT_FILE, "r")) {
        while (defined (my $line = <$fh>)) {
            if ($line =~ m|^cpu\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s|) {
                $res{user} = $1;
                $res{nice} = $2;
                $res{'system'} = $3;
                $res{idle} = $4;
                $res{used} = $1+$2+$3;
                $res{iowait} = $5;
            } elsif ($line =~ m|^cpu\d+\s|) {
                $cpucount++;
            }
        }
        $fh->close;
    }
    $cpucount = 1 if !$cpucount;

    # floating point time in seconds
    my $ctime = gettimeofday;
    $res{ctime} = $ctime;

    # Init last data
    %last_proc_stat = %res unless keys %last_proc_stat;

    # Calculate the diff time
    my $diff = ($ctime - $last_proc_stat{ctime}) * $clock_ticks * $cpucount;
    if ($diff > IO_DELAY_INTERVAL) { # don't update too often
        # CPU usage
        my $useddiff = $res{used} - $last_proc_stat{used};
           $useddiff = $diff if $useddiff > $diff;
        $res{cpuusage} = $useddiff/$diff;

        # IO delay
        my $waitdiff = $res{iowait} - $last_proc_stat{iowait};
           $waitdiff = $diff if $waitdiff > $diff;
        $res{iodelay} = $waitdiff/$diff;

        # Save it!
        %last_proc_stat = %res;
    } else {
        $res{cpuusage} = $last_proc_stat{cpuusage};
        $res{iodelay} = $last_proc_stat{iodelay};
    }

    return %res;
}
sub get_net_route {
    my @res;

    my $fh = IO::File->new(PROC_NET_ROUTE_FILE, "r");
    return () unless $fh;

    my $int_to_quad = sub {
       return join '.' => map { ($_[0] >> 8*(3-$_)) % 256 } (3, 2, 1, 0);
    };

    while (defined(my $line = <$fh>)) {
        next if $line =~/^Iface\s+Destination/; # skip head
        my ($iface, $dest, $gateway, $metric, $mask, $mtu) = (split(/\s+/, $line))[0,1,2,6,7,8];
        push @res, {
            destination => &$int_to_quad(hex($dest)),
            gateway     => &$int_to_quad(hex($gateway)),
            mask        => &$int_to_quad(hex($mask)),
            metric      => $metric,
            mtu         => $mtu,
            iface       => $iface,
        };
    }

    $fh->close();

    return @res;
}
sub get_net_ipv6_route {
    my @res;

    my $fh = IO::File->new(PROC_NET_IPV6_ROUTE_FILE, "r");
    return () unless $fh;

    my $read_v6addr = sub { $_[0] =~ s/....(?!$)/$&:/gr };

    # ipv6_route has no header
    while (defined(my $line = <$fh>)) {
        my ($dest, $prefix, $nexthop, $metric, $iface) = (split(/\s+/, $line))[0,1,4,5,9];
        push @res, {
            destination => &$read_v6addr($dest),
            prefix      => hex("$prefix"),
            gateway     => &$read_v6addr($nexthop),
            metric      => hex("$metric"),
            iface       => $iface
        };
    }

    $fh->close();

    return @res;
}
sub get_net_dev {
    my %res;

    my $fh = IO::File->new(PROC_NET_DEV_FILE, "r");
    return () unless $fh;

    while (defined (my $line = <$fh>)) {
        if ($line =~ m/^\s*(.*):\s*(\d+)\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+(\d+)\s+/) {
            $res{$1} = {
                receive  => $2 * 1,
                transmit => $3 * 1,
            };
        }
    }

    $fh->close();

    return %res;
}
sub get_mounts {
    my %res;

    # PROC_MOUNTS_FILE
    my $fh = IO::File->new(PROC_MOUNTS_FILE, "r");
    return () unless $fh;

    # mounts encode spaces (\040), tabs (\011), newlines (\012), backslashes (\\ or \134)
    my $decode_mount = sub {
        my $str = shift;
        #return $str =~ s/\\(?:040|01[12]|134|\\)/"\"$&\""/ger;
        $str =~ s/\\([0-7]{1,3})/chr(oct($1))/ge;
        return $str;
    };

    while (defined (my $line = <$fh>)) {
        if ($line =~/^\s*([^#].*)$/) {
            # lines from the file are encoded so we can just split at spaces
            my ($what, $dir, $fstype, $opts) = split(/[ \t]/, $1, 4);
            my ($freq, $passno) = (0, 0);

            # in glibc's parser frequency and pass seem to be optional
            $freq = $1 if $opts =~ s/\s+(\d+)$//;
            $passno = $1 if $opts =~ s/\s+(\d+)$//;
            my $w = &$decode_mount($what);
            $res{$w} = { # device
                    point   => &$decode_mount($dir),
                    fstype  => &$decode_mount($fstype),
                    opts    => &$decode_mount($opts),
                    freq    => $freq,
                    passno  => $passno,
                    # Extended fields from /proc/self/mountinfo
                    id      => 0, # unique identifier of the mount
                    majmin  => '0:0', # major and minor number of partitioned device
            };
        }
    }

    $fh->close();

    # PROC_MOUNTINFO_FILE
    #
    # 36  35  98:0 /mnt1 /mnt2 rw,noatime master:1 - ext3 /dev/root rw,errors=continue
    # (1) (2) (3)  (4)   (5)   (6)        (7)     (8) (9) (10)      (11)
    #
    # (1) mount ID: unique identifier of the mount (may be reused after umount)
    # (2) parent ID: ID of parent (or of self for the top of the mount tree)
    # (3) major:minor: value of st_dev for files on filesystem
    # (4) root: root of the mount within the filesystem
    # (5) mount point: mount point relative to the process's root
    # (6) mount options: per mount options
    # (7) optional fields: zero or more fields of the form "tag[:value]"
    # (8) separator: marks the end of the optional fields
    # (9) filesystem type: name of filesystem of the form "type[.subtype]"
    # (10) mount source: filesystem specific information or "none"
    # (11) super options: per super block options
    #
    if (my $fh = IO::File->new(PROC_MOUNTINFO_FILE, "r")) {
        while (defined (my $line = <$fh>)) {
            chomp $line;
            my @fields = split(/[ \t]/, $line);
            my $ign = pop @fields;
            my $w = &$decode_mount(pop(@fields) // '');
            next unless length($w);
            $res{$w}->{id} = shift(@fields) // 0;
            $ign = shift(@fields);
            $res{$w}->{majmin} = shift(@fields) // '0:0';
        }
        $fh->close();
    }

    return %res;
}
sub get_filesystems {
    my $nodev = shift || 0;
    my @res;

    my $fh = IO::File->new(PROC_FILESYSTEMS_FILE, "r");
    return () unless $fh;

    while (defined(my $line = <$fh>)) {
        my @rec = split(/\s+/, $line);
        my $filesystem = pop(@rec) // "";
        my $flag = pop(@rec) // "";
        push @res, $filesystem if $nodev || !($flag && $flag eq 'nodev');
    }

    $fh->close();

    return @res;
}
sub get_disk_free {
    my %mounts = get_mounts();
    my %disks = get_diskstats();
    my %blkio = get_blkio();
    my @fs = get_filesystems();
    foreach my $k (keys %mounts) {
        my $fstype = $mounts{$k}->{fstype};
        unless ($fstype && grep {$_ eq $fstype} @fs) {
            delete $mounts{$k};
            next;
        }

        # Set size by mountpoint
        my $point = $mounts{$k}->{point};
        my %du = df($point);
        $mounts{$k}->{$_} = $du{$_} for qw/avail free used total/;

        # Set IO data
        my $majmin = $mounts{$k}->{majmin} // '0:0';
        my ($maj, $min) = split(/\:/, $majmin, 2);
        my $majmin_root = sprintf('%d:0', $maj);
        my $rk = abs_path($k) // $k; # Resolve links
        my $blk;
        if ($blk = $blkio{$majmin}) {
            $mounts{$k}->{diskread} = $blk->{Read} || 0;
            $mounts{$k}->{diskwrite} = $blk->{Write} || 0;
        } elsif ($blk = $blkio{$majmin_root}) {
            $mounts{$k}->{diskread} = $blk->{Read} || 0;
            $mounts{$k}->{diskwrite} = $blk->{Write} || 0;
        } elsif ($disks{$rk}) {
            $mounts{$k}->{$_} = $disks{$rk}->{$_} for qw/diskread diskwrite/;
        } else {
            $mounts{$k}->{$_} = 0 for qw/diskread diskwrite/;
        }
    }

    return %mounts;
}
sub df {
    my $dir = shift // "/";
    my %res = (
        avail   => 0,
        free    => 0,
        used    => 0,
        total   => 0,
        bsize   => 0,
    );
    my $statvfs = _getstatvfs($dir);
    return %res if !defined($statvfs->{bsize}) || $statvfs->{bsize} == 0 || $statvfs->{blocks} == 0;

    # Init
    my ($blocks, $bfree, $bavail);
    $blocks = $statvfs->{blocks} // 0;
    $bfree = $statvfs->{bfree} // 0;
    $bavail = $statvfs->{bavail} // 0;

    # Check
    my $result = $statvfs->{bsize};
    $blocks *= $result;
    $bfree *= $result;
    if ($bavail < 0) {
        $result *= -1;
    }
    $bavail *= $result;

    %res = (
        avail   => $bavail,
        free    => $bfree,
        used    => $blocks - $bfree,
        total   => $blocks,
        bsize   => $result,
    );
    return %res;
}
sub get_diskstats {
    my %res;
    if (my $fh = IO::File->new(PROC_DISKSTATS_FILE, "r")) {
        while (defined (my $line = <$fh>)) {
            $line =~ s/^\s+//;
            my @rec = split(/\s+/, $line);
            next unless defined($rec[2]) && length($rec[2]);
            my $dev = File::Spec->catfile(DEV_DIR, $rec[2]);
            $res{$dev} = {
                name        => $rec[2],
                diskread    => $rec[5] || 0,
                diskwrite   => $rec[9] || 0,
            };
        }
        $fh->close();
    }
    return %res;
}
sub get_blkio {
    my %res;

    if (my $fh = IO::File->new(PROC_BLKIO_FILE, "r")) {
        while (defined (my $line = <$fh>)) {
            chomp $line;
            $line =~ s/^\s+//;
            next unless length($line);
            my ($majmin, $key, $val) = split(/[\t ]+/, $line, 3);
            $res{$majmin}{$key} = $val;
        }

        $fh->close();
    }
    return %res;
}
sub get_net_tcp {
    my $udp_proto = shift || 0;
    my @res;

    # This document describes the interfaces /proc/net/tcp and /proc/net/tcp6.
    # Note that these interfaces are deprecated in favor of tcp_diag.
    #
    # These /proc interfaces provide information about currently active TCP
    # connections, and are implemented by tcp4_seq_show() in net/ipv4/tcp_ipv4.c
    # and tcp6_seq_show() in net/ipv6/tcp_ipv6.c, respectively.
    #
    # It will first list all listening TCP sockets, and next list all established
    # TCP connections. A typical entry of /proc/net/tcp would look like this (split
    # up into 3 parts because of the length of the line):
    #
    #   46: 010310AC:9C4C 030310AC:1770 01
    #   |      |      |      |      |   |--> connection state
    #   |      |      |      |      |------> remote TCP port number
    #   |      |      |      |-------------> remote IPv4 address
    #   |      |      |--------------------> local TCP port number
    #   |      |---------------------------> local IPv4 address
    #   |----------------------------------> number of entry
    #
    #   00000150:00000000 01:00000019 00000000
    #      |        |     |     |       |--> number of unrecovered RTO timeouts
    #      |        |     |     |----------> number of jiffies until timer expires
    #      |        |     |----------------> timer_active (see below)
    #      |        |----------------------> receive-queue
    #      |-------------------------------> transmit-queue
    #
    #   1000        0 54165785 4 cd1e6040 25 4 27 3 -1
    #    |          |    |     |    |     |  | |  | |--> slow start size threshold,
    #    |          |    |     |    |     |  | |  |      or -1 if the threshold
    #    |          |    |     |    |     |  | |  |      is >= 0xFFFF
    #    |          |    |     |    |     |  | |  |----> sending congestion window
    #    |          |    |     |    |     |  | |-------> (ack.quick<<1)|ack.pingpong
    #    |          |    |     |    |     |  |---------> Predicted tick of soft clock
    #    |          |    |     |    |     |              (delayed ACK control data)
    #    |          |    |     |    |     |------------> retransmit timeout
    #    |          |    |     |    |------------------> location of socket in memory
    #    |          |    |     |-----------------------> socket reference count
    #    |          |    |-----------------------------> inode
    #    |          |----------------------------------> unanswered 0-window probes
    #    |---------------------------------------------> uid
    #
    # timer_active:
    #  0  no timer is pending
    #  1  retransmit-timer is pending
    #  2  another timer (e.g. delayed ack or keepalive) is pending
    #  3  this is a socket in TIME_WAIT state. Not all fields will contain
    #     data (or even exist)
    #  4  zero window probe timer is pending

    # Version 4
    my $f = $udp_proto ? PROC_NET_UDP : PROC_NET_TCP;
    return () unless -e $f;
    my $fh = IO::File->new($f, "r");
    if ($fh) {
        while (defined(my $line = <$fh>)) {
            $line =~ s/^\s+//;
            next if $line !~ /^[0-9]+/; # skip head
            my @row = split(/\s+/, $line);

            my $st_names = TCP_ST_NAMES;
            my $st2name = sub {
                my $st = shift;
                my $name = $st_names->[$st];
                return defined($name) ? $name : $st;
            };

            # Local IP:PORT
            my ($local_ip, $local_port) = ($row[1] =~ /([0-9A-F]+)\:([0-9A-F]+)/i) ? ($1, $2) : (0, 0);

            # Rem IP:PORT
            my ($rem_ip, $rem_port) = ($row[2] =~ /([0-9A-F]+)\:([0-9A-F]+)/i) ? ($1, $2) : (0, 0);

            push @res, {
                proto           => $udp_proto ? 'udp' : 'tcp',
                sl              => $row[0] =~ /(\d+)/ ? $1 : undef, # Get sl
                local_ip        => _hex2ip($local_ip),
                local_port      => hex($local_port),
                rem_ip          => _hex2ip($rem_ip),
                rem_port        => hex($rem_port),
                st              => &$st2name(hex($row[3])),
                uid             => $row[7],
                inode           => $row[9],
                user            => scalar(getpwuid($row[7] // 0)),
            };
        }
        $fh->close();
    }

    # Version 6
    my $f6 = $udp_proto ? PROC_NET_UDP6 : PROC_NET_TCP6;
    return @res unless -e $f6;
    my $fh6 = IO::File->new($f6, "r");
    if ($fh6) {
        while (defined(my $line = <$fh6>)) {
            $line =~ s/^\s+//;
            next if $line !~ /^[0-9]+/; # skip head
            my @row = split(/\s+/, $line);

            my $st_names = TCP_ST_NAMES;
            my $st2name = sub {
                my $st = shift;
                my $name = $st_names->[$st];
                return defined($name) ? $name : $st;
            };

            # Local IP:PORT
            my ($local_ip, $local_port) = ($row[1] =~ /([0-9A-F]+)\:([0-9A-F]+)/i) ? ($1, $2) : (0, 0);

            # Rem IP:PORT
            my ($rem_ip, $rem_port) = ($row[2] =~ /([0-9A-F]+)\:([0-9A-F]+)/i) ? ($1, $2) : (0, 0);

            push @res, {
                proto           => $udp_proto ? 'udp6' : 'tcp6',
                sl              => $row[0] =~ /(\d+)/ ? $1 : undef, # Get sl
                local_ip        => _hex2ip($local_ip),
                local_port      => hex($local_port),
                rem_ip          => _hex2ip($rem_ip),
                rem_port        => hex($rem_port),
                st              => &$st2name(hex($row[3])),
                uid             => $row[7],
                inode           => $row[9],
                user            => scalar(getpwuid($row[7] // 0)),
            };
        }
        $fh6->close();
    }

    return @res;
}
sub get_net_udp {
    get_net_tcp(1);
}

sub _hex2ip {
    my $h = shift;
    my $bin = pack "C*" => map hex, $h =~ /../g;
    my @l = unpack "L*", $bin;
    if (@l == 4) {
        return join ':', map { sprintf "%x:%x", $_ >> 16, $_ & 0xffff } @l;
    } elsif (@l == 1) {
        return join '.', map { $_ >> 24, ($_ >> 16 ) & 0xff, ($_ >> 8) & 0xff, $_ & 0xff } @l;
    }
    return 0;
}

1;

__END__
